import { IInterceptor, IKeyValueMap, IMapDidChange, IMapWillChange, Lambda } from "mobx";
import { ComplexType, IAnyType, IChildNodesMap, IComplexType, IContext, IJsonPatch, INode, IType, IValidationResult, ObjectNode, TypeFlags } from "../../internal";
export interface IMapType<C, S, T> extends IComplexType<IKeyValueMap<C> | undefined, IKeyValueMap<S>, IMSTMap<C, S, T>> {
    flags: TypeFlags.Optional;
}
export interface IMSTMap<C, S, T> {
    clear(): void;
    delete(key: string): boolean;
    forEach(callbackfn: (value: T, key: string, map: IMSTMap<C, S, T>) => void, thisArg?: any): void;
    get(key: string): T | undefined;
    has(key: string): boolean;
    set(key: string, value: C | S | T): this;
    readonly size: number;
    put(value: C | S | T): T;
    keys(): IterableIterator<string>;
    values(): IterableIterator<T>;
    entries(): IterableIterator<[string, T]>;
    [Symbol.iterator](): IterableIterator<[string, T]>;
    /** Merge another object into this map, returns self. */
    merge(other: IMSTMap<any, any, T> | IKeyValueMap<C | S | T> | any): this;
    clear(): void;
    replace(values: IMSTMap<any, any, T> | IKeyValueMap<T>): this;
    /**
     * Returns a plain object that represents this map.
     * Note that all the keys being stringified.
     * If there are duplicating keys after converting them to strings, behaviour is undetermined.
     */
    toPOJO(): IKeyValueMap<T>;
    /**
     * Returns a shallow non observable object clone of this map.
     * Note that the values migth still be observable. For a deep clone use mobx.toJS.
     */
    toJS(): Map<string, T>;
    toJSON(): IKeyValueMap<T>;
    toString(): string;
    [Symbol.toStringTag]: "Map";
    /**
     * Observes this object. Triggers for the events 'add', 'update' and 'delete'.
     * See: https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/observe
     * for callback details
     */
    observe(listener: (changes: IMapDidChange<string, T>) => void, fireImmediately?: boolean): Lambda;
    intercept(handler: IInterceptor<IMapWillChange<string, T>>): Lambda;
}
export declare enum MapIdentifierMode {
    UNKNOWN = 0,
    YES = 1,
    NO = 2
}
export declare class MapType<C, S, T> extends ComplexType<IKeyValueMap<C> | undefined, IKeyValueMap<S>, IMSTMap<C, S, T>> {
    shouldAttachNode: boolean;
    subType: IAnyType;
    identifierMode: MapIdentifierMode;
    identifierAttribute: string | undefined;
    readonly flags: TypeFlags;
    constructor(name: string, subType: IAnyType);
    instantiate(parent: ObjectNode | null, subpath: string, environment: any, snapshot: S): INode;
    private _determineIdentifierMode;
    initializeChildNodes(objNode: ObjectNode, initialSnapshot?: any): IChildNodesMap;
    createNewInstance(node: INode, childNodes: IChildNodesMap, snapshot: any): IMSTMap<any, any, any>;
    finalizeNewInstance(node: ObjectNode, instance: any): void;
    describe(): string;
    getChildren(node: ObjectNode): ReadonlyArray<INode>;
    getChildNode(node: ObjectNode, key: string): INode;
    willChange(change: IMapWillChange<any, any>): IMapWillChange<any, any> | null;
    private processIdentifier;
    getValue(node: ObjectNode): any;
    getSnapshot(node: ObjectNode): IKeyValueMap<S>;
    processInitialSnapshot(childNodes: IChildNodesMap, snapshot: any): any;
    didChange(change: IMapDidChange<any, any>): void;
    applyPatchLocally(node: ObjectNode, subpath: string, patch: IJsonPatch): void;
    applySnapshot(node: ObjectNode, snapshot: any): void;
    getChildType(key: string): IAnyType;
    isValidSnapshot(value: any, context: IContext): IValidationResult;
    getDefaultSnapshot(): {};
    removeChild(node: ObjectNode, subpath: string): void;
}
/**
 * Creates a key based collection type who's children are all of a uniform declared type.
 * If the type stored in a map has an identifier, it is mandatory to store the child under that identifier in the map.
 *
 * This type will always produce [observable maps](https://mobx.js.org/refguide/map.html)
 *
 * @example
 * const Todo = types.model({
 *   id: types.identifier,
 *   task: types.string
 * })
 *
 * const TodoStore = types.model({
 *   todos: types.map(Todo)
 * })
 *
 * const s = TodoStore.create({ todos: {} })
 * unprotect(s)
 * s.todos.set(17, { task: "Grab coffee", id: 17 })
 * s.todos.put({ task: "Grab cookie", id: 18 }) // put will infer key from the identifier
 * console.log(s.todos.get(17).task) // prints: "Grab coffee"
 *
 * @export
 * @alias types.map
 * @param {IType<S, T>} subtype
 * @returns {IComplexType<S[], IObservableArray<T>>}
 */
export declare function map<C, S, T>(subtype: IType<C, S, T>): IMapType<C, S, T>;
export declare function isMapType<IT extends IMapType<any, any, any>>(type: IT): type is IT;
